// awuiobj.c - implement class AWuiObject
// Copyright (c) 1994 by Ron Burk

#include <string.h>
#include <stdio.h>

#include "wuidbint.h"
#include "wuierror.h"
#include "wuistd.h"
#include "wuiobjdb.h"



/////////////////////////////////////////////////
// AWuiObject - supply minimal object functions

AWuiObject     *AWuiObject::Root            = 0;
int             AWuiObject::TraceFlag       = 1;
TWuiDatabase   *AWuiObject::Database        = 0;
TWuiObjects    *AWuiObject::MasterObjects   = 0;
TWuiObjects    *AWuiObject::Methods         = 0;


TWuiNameBase   AWuiObject::M_NASTY_POINTER;

TWuiNameBase AWuiObject::M_CLONE;
TWuiNameBase AWuiObject::M_OBJECTSAVE;
TWuiNameBase AWuiObject::M_OBJECTDUMP;
TWuiNameBase AWuiObject::M_DELETE;
TWuiNameBase AWuiObject::M_ISA;
TWuiNameBase AWuiObject::M_RENAME;
TWuiNameBase AWuiObject::M_CLASS;
TWuiNameBase AWuiObject::M_PARENT;
TWuiNameBase AWuiObject::M_CHILD;
TWuiNameBase AWuiObject::M_ATTRIBUTE;
TWuiNameBase AWuiObject::M_EDITATTR;

TWuiNameBase AWuiObject::P_PERSISTENT;
TWuiNameBase AWuiObject::P_INHERITED;
TWuiNameBase AWuiObject::P_INHERITABLE;
TWuiNameBase AWuiObject::P_READONLY;
TWuiNameBase AWuiObject::P_REFERENCE;

TWuiIntPropertyBit *AWuiObject::P_Persistent;
TWuiIntPropertyBit *AWuiObject::P_Inherited;
TWuiIntPropertyBit *AWuiObject::P_Inheritable;
TWuiIntPropertyBit *AWuiObject::P_ReadOnly;

TWuiObjects *AWuiObject::PropertyProperties;

TWuiNameBase AWuiObject::P_CLASS;
TWuiNameBase AWuiObject::P_PARENTCLASS;
TWuiNameBase AWuiObject::P_COMPILER;
TWuiNameBase AWuiObject::P_WUIMANVERSION;


TWuiNameBase AWuiObject::S_CLASSES;
TWuiNameBase AWuiObject::S_ROOT;

TWuiNameBase AWuiObject::S_INHERITED;
TWuiNameBase AWuiObject::S_INHERITABLE;
TWuiNameBase AWuiObject::S_PERSISTENT;
TWuiNameBase AWuiObject::S_READONLY;

TWuiNameBase AWuiObject::C_ROOT;
TWuiNameBase AWuiObject::C_TOBJECT;
TWuiNameBase AWuiObject::C_TNONCONTAINER;
TWuiNameBase AWuiObject::C_TCONTAINER;
TWuiNameBase AWuiObject::C_TROOT;
TWuiNameBase AWuiObject::C_TATTRIBUTE;
TWuiNameBase AWuiObject::C_TATTRIBUTEMETHOD;
TWuiNameBase AWuiObject::C_TATTRIBUTESTRING;
TWuiNameBase AWuiObject::C_TATTRIBUTEINT;
TWuiNameBase AWuiObject::C_TLONGPROPERTY;
TWuiNameBase AWuiObject::C_TATTRIBUTEDOUBLE;
TWuiNameBase AWuiObject::C_TATTRIBUTEINTERNALBIT;


void AWuiObject::ClassDestructor(const TWuiName /*Name*/)
    {
    // ??? do deletes
    }

AWuiObject *AWuiObject::ClassConstructor(const TWuiName /*Name*/)
    {
    new (&M_NASTY_POINTER) TWuiName("M_NASTY_POINTER");

    new (&M_CLONE) TWuiName("m_Clone");
    new (&M_OBJECTSAVE) TWuiName("m_ObjectSave");
    new (&M_OBJECTDUMP) TWuiName("m_ObjectDump");
    new (&M_DELETE) TWuiName("m_Delete");
    new (&M_ISA) TWuiName("m_IsA");
    new (&M_RENAME) TWuiName("m_Rename");
    new (&M_CLASS) TWuiName("m_Class");
    new (&M_PARENT) TWuiName("m_ParentClass");
    new (&M_CHILD) TWuiName("m_Child");
    new (&M_ATTRIBUTE) TWuiName("m_Attribute");
    new (&M_EDITATTR) TWuiName("m_AttributeEdit");
    new (&P_PERSISTENT) TWuiName("p_Persistent");
    new (&P_INHERITED) TWuiName("p_Inherited");
    new (&P_INHERITABLE) TWuiName("p_Inheritable");
    new (&P_READONLY) TWuiName("p_ReadOnly");
    new (&P_REFERENCE) TWuiName("p_Reference");
    new (&P_CLASS) TWuiName("p_Class");
    new (&P_PARENTCLASS) TWuiName("p_ParentClass");
    new (&P_COMPILER) TWuiName("p_Compiler");
    new (&P_WUIMANVERSION) TWuiName("p_WuimanVersion");
    new (&S_CLASSES) TWuiName("Classes");
    new (&S_ROOT) TWuiName("/");
    new (&S_INHERITED) TWuiName("Inherited");
    new (&S_INHERITABLE) TWuiName("Inheritable");
    new (&S_PERSISTENT) TWuiName("Persistent");
    new (&S_READONLY) TWuiName("ReadOnly");
    new (&C_ROOT) TWuiName("/");
    new (&C_TOBJECT) TWuiName("TObject");
    new (&C_TNONCONTAINER) TWuiName("TNonContainer");
    new (&C_TCONTAINER) TWuiName("TContainer");
    new (&C_TROOT) TWuiName("/");
    new (&C_TATTRIBUTE) TWuiName("p_");
    new (&C_TATTRIBUTEMETHOD) TWuiName("m_");
    new (&C_TATTRIBUTESTRING) TWuiName("p_String");
    new (&C_TATTRIBUTEINT) TWuiName("p_Int");
    new (&C_TLONGPROPERTY) TWuiName("p_Long");
    new (&C_TATTRIBUTEDOUBLE) TWuiName("p_Double");
    new (&C_TATTRIBUTEINTERNALBIT) TWuiName("p_InternalBit");


    P_Persistent    = new TWuiIntPropertyBit(P_PERSISTENT);
    P_Inherited     = new TWuiIntPropertyBit(P_INHERITED);
    P_Inheritable   = new TWuiIntPropertyBit(P_INHERITABLE);
    P_ReadOnly      = new TWuiIntPropertyBit(P_READONLY);

    PropertyProperties = new TWuiObjects(4);
    PropertyProperties->Insert(P_Persistent);
    PropertyProperties->Insert(P_Inherited);
    PropertyProperties->Insert(P_Inheritable);
    PropertyProperties->Insert(P_ReadOnly);
    Methods     = new TWuiObjects(32);
    return NULL;
    }

AWuiObject::AWuiObject(const TWuiName ObjectName)
    :   Name_(ObjectName),
                    Flags_(PERSISTENT|INHERITABLE|INHERITED)
    {   MEMBERASSERT();
    }


// WUIMAN Object destructor - delete objects in list.
AWuiObject::~AWuiObject()
    {   MEMBERASSERT();
    AWuiObject  *Object = 0;

    TWuiObjects *List = Children();
    if(List)
        {
        while((Object = List->Remove()) != NULL)
            if(Object->GetFlags(REFERENCE) == 0)
                delete Object;
        delete List;
        }

    List    = Attributes();
    if(List)
        {
        while((Object = List->Remove()) != NULL)
            if(Object->GetFlags(REFERENCE) == 0)
                delete Object;
        delete List;
        }
    }

long    AWuiObject::Rename(const TWuiName NewName)
    {   MEMBERASSERT();
    Name_   = NewName;
    return TRUE;
    }

// Save() - base-level save function.
//          
long    AWuiObject::Save(TWuiPath Path, int Depth) const
    {   MEMBERASSERT();
DEBUG_TRACE(">AWuiObject::Save< '%s', '%s'",
    Path.AbsolutePath(), Path.RelativePath());

    long    Result = TRUE;
    long    Status;
    // if not persistent, we don't save
    // (and none of our children save)
    if(GetFlags(PERSISTENT) == 0)
        return FALSE;
    TWuiObjects *List = Attributes();
    if(List != NULL)
        {
        for(int i=0; i < List->NObjects(); ++i)
            {
            AWuiObject *Object = List->Get(i);
            ASSERT(Object != NULL);
            if(!Object->GetFlags(PERSISTENT))
                continue;
            char    Name[TWuiName::MAXNAME+2];
            Name[0] = '.';
            Object->Name().GetName(Name+1);
            // must be property and not built-in property bit
            if(Name[1] != 'p' || NameToFlag(Object->Name()))
                continue;
            Path.Push(Name);
            Status = Object->Save(Path, 1);
            Path.Pop();
            if(Status < WUIMAN_ERRORS)
                Result  = Status;
            }
        }
    if(Depth <= 0)
        return FALSE;
    List    = Children();
    if(List != NULL)
        {
        for(int i=0; i < List->NObjects(); ++i)
            {
            AWuiObject *Object = List->Get(i);
            ASSERT(Object != NULL);
            if(!Object->GetFlags(PERSISTENT))
                continue;
            Path.Push(Object->Name());
            Status  = Object->Save(Path, Depth-1);
            Path.Pop();
            if(Status < WUIMAN_ERRORS)
                Result  = Status;
            }
        }
    return Result;
    }

int AWuiObject::SetFlags(int Bits, int On)
    {   MEMBERASSERT();
    if(On)
        return Flags_  |= Bits;
    else
        return ClearFlags(Bits);
    }
int AWuiObject::ClearFlags(int Bits)
    {   MEMBERASSERT();

    return Flags_  &= ~Bits;
    }
int AWuiObject::GetFlags(int Bits) const
    {   MEMBERASSERT();
    return (Flags_ & Bits);
    }
int AWuiObject::NameToFlag(const TWuiName Name) const
    {   MEMBERASSERT();
    if(Name == P_PERSISTENT)
        return PERSISTENT;
    else if(Name == P_INHERITED)
        return INHERITED;
    else if(Name == P_READONLY)
        return READONLY;
    else if(Name == P_INHERITABLE)
        return INHERITABLE;
    else if(Name == P_REFERENCE)
        return REFERENCE;
    else
        return 0;
    }


void    AWuiObject::DumpObject()
    {   MEMBERASSERT();
    char    Buffer[TWuiName::MAXNAME];
    TWuiObjects *List = Children();
    AWuiObject *Item;
    int i;
    DEBUG_TRACE("**Children of '%s'**", Name().GetName(Buffer));
    int n = 0;
    if(List)
        {
        n   = List->NObjects();
        for(i=0; i < n; ++i)
            {
            Item = List->Get(i);
            Item->Name().GetName(Buffer);
            DEBUG_MORE("    '%s'", Buffer);
            }
        }
    List    = Attributes();
    DEBUG_TRACE("**Attributes**");
    if(List)
        {
        char    Class[TWuiName::MAXNAME], Parent[TWuiName::MAXNAME];
        n = List->NObjects();
        for(i=0; i < n; ++i)
            {
            Item = List->Get(i);
            Item->Name().GetName(Buffer);
            Item->ClassName().GetName(Class);
            Item->ParentClassName().GetName(Parent);
            DEBUG_MORE("    '%s'   (class='%s', parent='%s')",
                        Buffer, Class, Parent);
            }
        }
    }

void    AWuiObject::Trace(const char *Function,
                              TWuiPath Path, TWuiArgs *Args)
    {   MEMBERASSERT();
    char    AttributeName[TWuiName::MAXNAME + 256];
    char    MyName[TWuiName::MAXNAME];
    if(Args != 0)
        {
        Args->AttributeName().GetName(AttributeName);
        strcat(AttributeName, "(");
        for(int i = 0; i < Args->NArgs(); ++i)
            {
            if(i)
                strcat(AttributeName, ",");
            if(Args->GetType(i) == TWuiArgs::LONG_ARG)
                {
                char    Buffer[32];
                sprintf(Buffer, "%ld", Args->GetLong(i));
                strcat(AttributeName, Buffer);
                }
            else if(Args->GetType(i) == TWuiArgs::STRING_ARG)
                strcat(AttributeName, Args->GetString(i));
            else
                strcat(AttributeName, "???");
            }
        strcat(AttributeName, ")");
        }
    else
        strcpy(AttributeName, "???");
    this->Name().GetName(MyName);
    const char *AbsolutePath = Path.AbsolutePath()+1;
    for(int Count = 0; AbsolutePath != NULL
                && AbsolutePath < Path.RelativePath(); ++Count)
        AbsolutePath = strchr(AbsolutePath+1, '/');
    Count   *= 4;
    DEBUG_TRACE("%*.*s[%s][%s] AbsPath = '%s', RelPath = '%s'"
                " Attribute='%s'",
                    Count, Count, "",
                                            Function, MyName,
                    Path.AbsolutePath(), Path.RelativePath(),
                    AttributeName);
    }


// Get - default get attribute handler
//      NOTE: the default is to forward the message, or to return
//            FALSE if this object is the destination.
long    AWuiObject::Get(TWuiPath Path, TWuiArgs *Args,
                             char *Buffer, size_t MaxLength)
    {   MEMBERASSERT();
    ASSERT(Args != NULL);
    if(AWuiObject::TraceFlag)
        Trace("AWuiObject::Get", Path, Args);
    TWuiObjects *MyAttributes = Attributes();
    const TWuiName AttributeName= Args->AttributeName();
    long    Result  = TRUE; // assume success
    if(Path.Descend())      // if message is for a child node
        {
        // then try to locate child and pass it message
        TWuiName ChildName      = Path.CurrentName();
        TWuiObjects *MyChildren;
        // caller can send message directly to attribute
        // by using path that ends in ".attrname".
        if(Path.RelativePath()[0] == '.')
            MyChildren = Attributes();
        else
            MyChildren = Children();
        AWuiObject  *Child      = NULL;
        if(MyChildren)
            Child   = MyChildren->Get(ChildName);
        if(Child)
            {
            // pass message on to child named in path
            Result  = Child->Get(Path, Args, Buffer, MaxLength);
            // just return any errors
            if(Result < WUIMAN_ERRORS)
                return Result;
            if(AttributeName == M_DELETE)
                {
                // if message was for my child
                if(Path.Descend() == FALSE)
                    {
                    Database->DeletePath(Path.AbsolutePath());
                    delete MyChildren->Remove(ChildName);
                    }
                }
            }
        else    // else, no child by that name!
            {
            char    Name[TWuiName::MAXNAME];
            ChildName.GetName(Name);
            DEBUG_ERROR("[AWuiObject::Get]'%s' no such child", Name);
            DumpObject();
            Result  = WUIMAN_ERR_NO_SUCH_OBJECT;
            }
        }
    else        // else message is for me
        {
        TWuiName AttributeName = Args->AttributeName();

        // M_CHILD() returns # of children
        // M_CHILD(n) returns name of nth child
        // M_CHILD('name') returns TRUE if it exists
        if(AttributeName == M_CHILD
            || AttributeName == M_ATTRIBUTE)
            {
            TWuiObjects *MyChildren;
            if(AttributeName == M_ATTRIBUTE)
                MyChildren  = MyAttributes;
            else
                MyChildren  = Children();
            AWuiObject  *Child = 0;
            // m_xxxx() == # of children/attributes
            if(Args->NArgs() <= 0)
                return MyChildren
                             ? MyChildren->NObjects() : 0;
            // m_xxxx(n) == return nth child/attribute
            else if(Args->NArgs() == 1 && MyChildren)
                {
                if(Args->GetType(0) == TWuiArgs::LONG_ARG)
                    {
                    long    ArgNumber = Args->GetLong(0);
                    if(ArgNumber >= 0 && ArgNumber
                                 < MyChildren->NObjects())
                        {
                        Child = MyChildren
                                    ->Get(int(ArgNumber));
                        Child->Name().GetName(Buffer, MaxLength);
                        return 0;
                        }
                    }
                else if(Args->GetType(0) == TWuiArgs::STRING_ARG)
                    {
                    return MyChildren->Get(Args->GetString(0)) != NULL;
                    }
                else // else attr arg is wrong type
                    return WUIMAN_ERR_BAD_ATTR_ARG;
                return WUIMAN_ERR_BAD_ATTR_ARG;
                }
            else if(AttributeName == M_CHILD)
                return WUIMAN_ERR_BAD_ATTR_ARG;
            // m_xxxx('attr','prop') == return property bit
            else if(Args->NArgs() == 2 && MyAttributes
                && Args->GetType(0) == TWuiArgs::STRING_ARG)
                {
                Child = MyAttributes->Get(
                                        Args->GetString(0));
                if(Child)
                    return Child->Get(Path, Args,
                              Buffer, MaxLength);
                else
                    {
                    DEBUG_ERROR("No such attribute '%s'.", Args->GetString(0));
                    DumpObject();
                    return WUIMAN_ERR_BAD_ATTR;
                    }
                }
            }
        else if(AttributeName == M_CLASS)
            {
            // ??? need error checking ???
            ClassName().GetName(Buffer, MaxLength);
            return strlen(Buffer);
            }
        else if(AttributeName == M_PARENT)
            {
            // ??? need error checking ???
            ParentClassName().GetName(Buffer, MaxLength);
            return strlen(Buffer);
            }
        else if(AttributeName == M_ISA)
            {
            int NArgs   = Args->NArgs();
            if(NArgs < 1 || NArgs > 2
                || Args->GetType(0) != TWuiArgs::STRING_ARG)
                return  WUIMAN_ERR_BAD_ATTR_ARG;
            if(NArgs > 1 && Args->GetType(1) != TWuiArgs::STRING_ARG)
                return  WUIMAN_ERR_BAD_ATTR_ARG;
            if(NArgs > 1)
                return IsA(Args->GetString(0), Args->GetString(1));
            else
                return IsA(Args->GetString(0));
            }
        else if(AttributeName == M_OBJECTDUMP)
            {
            DumpObject();
            return TRUE;
            }
        // internal nasty hack to obtain pointer to node
        else if(AttributeName == M_NASTY_POINTER)
            {
            AWuiObject  *This = this;
            memcpy(Buffer, &This, sizeof(This));
            return 0;
            }
        // else if cloning me and mine
        else if(AttributeName == M_CLONE)
            {
            if( Args->NArgs() != 3
             || Args->GetType(0) != TWuiArgs::LONG_ARG
             || Args->GetType(1) != TWuiArgs::STRING_ARG
             || Args->GetType(2) != TWuiArgs::STRING_ARG
              )
                return WUIMAN_ERR_BAD_ATTR_ARG;
            long    Depth   = Args->GetLong(0);
            char   *ToPath  = Args->GetString(1);
            char   *NewName = Args->GetString(2);
				return Clone(int(Depth), ToPath, NewName);
            }
        else if(AttributeName == M_OBJECTDUMP)
            //  m_ObjectDump
            {
            DumpObject();
            return 0;
            }
        else if(AttributeName == M_OBJECTSAVE)
            //  m_Save([Depth])
            {
            int     Depth   = INT_MAX;
            int     NArgs   = Args->NArgs();
            if(NArgs > 1)
                return WUIMAN_ERR_BAD_ATTR_ARG;
            else if(NArgs > 0)
                {
                if(Args->GetType(0) != TWuiArgs::LONG_ARG)
                    return WUIMAN_ERR_BAD_ATTR_ARG;
                long Value = Args->GetLong(0);
                if(Value > INT_MAX)
                    Value   = INT_MAX;
                Depth   = int(Value);
                }
            return Save(Path, Depth);
            }
        else if(MyAttributes)
            {
            AWuiObject  *Target =
                   MyAttributes->Get(AttributeName);
            if(Target)
                {
                int Flag  = NameToFlag(AttributeName);
                if(Flag)
                    {
                    Flag = GetFlags(Flag);
                    if(Buffer != NULL)
                        {
                        if(MaxLength > 1)
                            *Buffer++ = (Flag ? '1' : '0');
                        if(MaxLength > 0)
                            *Buffer++ = '\0';
                        }
                    return Flag != 0;
                    }
                // else, it is not an internal attribute
                else
                    return Target->Get(Path, Args, Buffer, MaxLength);
                }
            }
        Result  = WUIMAN_ERR_BAD_ATTR;
        char    Name[TWuiName::MAXNAME+1];
        AttributeName.GetName(Name);
        DEBUG_WARNING("'%s': No such attribute for path '%s'.",
                    Name, Path.AbsolutePath());
        DumpObject();
        }
    return Result;
    }


/*  Clone - Clone a sub-tree to desired depth
 *  
 *  
 */
long    AWuiObject::Clone(int Depth, char *ToPath,
                                     const TWuiName NewName)
    {   MEMBERASSERT();
    if(Depth <= 0)
        return 0;
    // first, obtain pointer to destination object
    TWuiArgs    Args(M_NASTY_POINTER);
    ASSERT(Root != NULL);
    TCharBuffer DestPath(256);
    strcpy(DestPath, ToPath);
    AWuiObject *Destination = 0;
    long Status = Root->Get((char *)DestPath, &Args,
                 (char *)&Destination, sizeof(Destination));
    if(Status < WUIMAN_ERRORS)
        return Status;
    ASSERT(Destination != NULL);
    // now we have a pointer to the prospective
    // parent of our clone
    int Persistent  = Destination->GetFlags(PERSISTENT);

    int IsAttribute = IsA(C_TATTRIBUTE);
    TWuiObjects *List = IsAttribute ? Destination->Attributes()
                                    : Destination->Children();
    // check whether supports attributes/children
    if(List == NULL)
        return IsAttribute ? WUIMAN_ERR_TARGET_HATES_ATTR
                           : WUIMAN_ERR_TARGET_HATES_KIDS;
    // check for name collision
    if(List->Get(NewName))
        return WUIMAN_ERR_DUPLICATE_NAME;
    // clone ourselves
    AWuiObject *Subtree = Clone(NewName, Depth);
    List->Insert(Subtree);
    if(Persistent)
        Destination->Save((char *)DestPath, Depth);
    return 1;
    }


// Set - default Set attribute handler
//      NOTE: the default is to forward the message, or to return
//            FALSE if this object is the destination
long    AWuiObject::Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("AWuiObject::Set", Path, Args);
    TWuiObjects *MyAttributes   = Attributes();
    TWuiObjects *MyChildren     = Children();
    const TWuiName AttributeName= Args->AttributeName();
    long    Result  = TRUE; // assume success
    if(Path.Descend())      // if message is for a child
        {
        TWuiName ChildName      = Path.CurrentName();
        if(Path.RelativePath()[0] == '.')
            MyChildren = Attributes();
        else
            MyChildren = Children();
        AWuiObject  *Child      = NULL;
        if(MyChildren)
            Child   = MyChildren->Get(ChildName);
        if(Child)
            {
            // pass message on to child named in path
            Result  = Child->Set(Path, Args, Buffer);
            // just return any errors
            if(Result < WUIMAN_ERRORS)
                return Result;
            if(AttributeName == M_DELETE)
                {
                // if message was for my child
                if(Path.Descend() == FALSE)
                    delete MyChildren->Remove(ChildName);
                }
            else if(AttributeName == P_PERSISTENT)
                {

                // something was made persistent, so
                // everything above it must be too
                if(!GetFlags(PERSISTENT))
                    {
                    SetFlags(PERSISTENT);
DEBUG_TRACE("AWuiObject::Set() calls Save function!");
                    Save(Path, 0);
                    }
                }
            }
        else    // else, no child by that name!
            {
            char Buffer[TWuiName::MAXNAME];
            DEBUG_ERROR("Child '%s' not found in AWuiObject::Set()",
                ChildName.GetName(Buffer));
            DumpObject();
            Result  = WUIMAN_ERR_NO_SUCH_OBJECT;
            }
        }
    else        // else message is for me
        {
        if(!MyAttributes)
            {
            DumpObject();
            return WUIMAN_ERR_BAD_ATTR;
            }
        AWuiModifiableAttribute *Target = (AWuiModifiableAttribute *)
                        MyAttributes->Get(AttributeName);
        if(Target == NULL)
            {
            DumpObject();
            return WUIMAN_ERR_BAD_ATTR;
            }
        if(AttributeName == M_DELETE)
            return TRUE;    // parent will delete
        else if(AttributeName == M_RENAME)
            return Rename(Buffer);
        else    // else, caller is simply setting attribute value
            {
            // first, handle built-in attributes
            int State = (atoi(Buffer) != 0);
            int Flag  = NameToFlag(AttributeName);
            if(Flag)
                {
                SetFlags(Flag, State);
                Result = Target->Set(Path, Args, Buffer);
                if(Result >= WUIMAN_ERRORS)
                    {
                    if(Flag == PERSISTENT)
                        if(State == TRUE)
                            Result = Save(Path, 0);
                        else
                            ;   //??? delete path ???
                    }
                return Result;
                }
            // else, it is not an internal attribute
            Result = Target->Set(Path, Args, Buffer);
            if(Result < WUIMAN_ERRORS)
                return Result;
            // implement value inheritance!
#ifdef _MSC_VER
            if(((AWuiObject *)Target)->GetFlags(INHERITABLE)
                && MyChildren != NULL)
#else
            if(Target->GetFlags(INHERITABLE)
                && MyChildren != NULL)
#endif
                {
                AWuiObject  *Child = 0;
                // for each child node
                int NChildren = MyChildren->NObjects();
                for(int i=0; i < NChildren; ++i)
                    {
                    Child = MyChildren->Get(i);
                    ASSERT(Child != NULL);
                    TWuiObjects *AttrList = Target->Attributes();
                    // does it have attributes?
                    if(AttrList == NULL)
                        continue;
                    AWuiModifiableAttribute *Attr = (AWuiModifiableAttribute *)AttrList->Get(Name());
                    // does it have an attribute with my name?
                    if(Attr == NULL)
                        continue;
                    // is it of compatible type?
                    if(Attr->IsA(Target->ClassName()) == FALSE)
                        continue;
                    // does it wish to inherit?
#ifdef _MSC_VER
                    if(((AWuiObject *)Attr)->GetFlags(INHERITED) == 0)
                        continue;
#else
                    if(Attr->GetFlags(INHERITED) == 0)
                        continue;
#endif
                    Path.Push(Child->Name());
                    Child->Set(Path, Args, Buffer);
#ifdef _MSC_VER
                    ((AWuiObject *)Attr)->SetFlags(INHERITED);
#else
                    Attr->SetFlags(INHERITED);
#endif
                    Path.Pop();
                    }
                }
            }
        }
    return Result;
    }



// Broadcast - Broadcast set or get message to all objects in a list.
//
// the function returns TRUE if there were no errors. Otherwise,
// it returns the most recent error (does not stop if a particular
// object returns an error).
long    AWuiObject::Broadcast(TWuiObjects *List,
                              TWuiPath Path, TWuiArgs *Args,
              const char *Buffer, size_t MaxLength, int Set)
    {   MEMBERASSERT();
    long    Result = TRUE;
    if(List)
        {
        char   *Prefix  = NULL;
        long    Status  = Result;
        for(int i=0; i < List->NObjects(); ++i)
            {
            AWuiObject *Object = List->Get(i);
            ASSERT(Object != NULL);
            if(Prefix == NULL)
                {
                if(Object->IsA(C_TATTRIBUTE))
                    Prefix  = ".";
                else
                    Prefix  = "";
                }
            char    Name[TWuiName::MAXNAME+2];
            strcpy(Name, Prefix);
            Object->Name().GetName(Name + strlen(Name));
            Path.Push(Name);
            if(Set != 0)
                Status = Object->Get(Path, Args, (char *)Buffer, MaxLength);
            else
                Status = Object->Set(Path, Args, Buffer);
            Path.Pop();
            if(Status < WUIMAN_ERRORS)
                Result  = Status;
            }
        }
    return Result;
    }


TWuiObject *AWuiObject::NewObject(const TWuiName Type,
                const TWuiName Name)
    {
    AWuiObject *Result = AWuiObject::Clone(Type, Name);

    // assert that it is really an object (not an attribute)
    if(Result != 0)
        ASSERT(Result->IsA(C_TOBJECT));

    return (TWuiContainer *)Result;
    }

AWuiAttribute *AWuiObject::NewAttribute(const TWuiName Type,
                const TWuiName Name)
    {
#if 1
char Buffer1[TWuiName::MAXNAME];
char Buffer2[TWuiName::MAXNAME];
Type.GetName(Buffer1);
Name.GetName(Buffer2);
DEBUG_TRACE("NewAttribute(type='%s', name='%s')", Buffer1, Buffer2);
#endif

    AWuiObject *Result = AWuiObject::Clone(Type, Name);

    // assert that it is really an attribute
    if(Result != 0)
        ASSERT(Result->IsA(C_TATTRIBUTE));

    return (AWuiAttribute *)Result;
    }

long        AWuiObject::Load(TWuiPath Path, const char * /*Value*/)
    {   MEMBERASSERT();
    ASSERTDATAPTR(AWuiObject::Database);
    long Status = 0;
    TWuiDatabaseKeys *Keys  = AWuiObject::Database->
                               GetKeys(Path.AbsolutePath());
    if(Keys == NULL)    // if no attributes or children
        return 0;
    const int MAX_ITEM = 1024;
    TCharBuffer NodeData(MAX_ITEM);
    // first, load the attributes, creating when necessary.
    const char *Name = Keys->FirstKey();
    ASSERTSTRINGPTR(Name);
DEBUG_TRACE("::Load First attribute name '%s'.", Name);
    TWuiObjects *List = Attributes();
    if(List)    // if object supports attributes
        for(; Name != NULL; Name = Keys->NextKey())
            {
            // Read class name, value string
            int ReadStatus = AWuiObject::Database->ReadItem(
                                  Path.AbsolutePath(), Name,
                                      (char *)NodeData, MAX_ITEM);
            if(ReadStatus == -1)
                continue;
            if(*Name++ != '.')  // if not attribute
                continue;
DEBUG_TRACE("::Load Just read attribute '%s'='%s'", Name, (char *)NodeData);
            char *Value = strchr(NodeData, ',');
            if(Value)
                *Value++    = '\0';
            // if built-in attribute bit
            int Bit = NameToFlag(Name);
            if(Bit != 0)
                {
                SetFlags(Bit);
                continue;
                }
            AWuiAttribute *Attribute = (AWuiAttribute *)
                                            List->Get(Name);
            // if object did not create this attribute
            if(!Attribute)
                { // then create it ourselves
DEBUG_TRACE("::Load Create the attribute");
                Attribute   = AWuiObject::NewAttribute(
                                  TWuiName(NodeData), Name);
                if(Attribute)
                    List->Insert(Attribute);
                else
                    continue;
                }
            Path.Push(--Name);
            Status  = Attribute->Load(Path, Value);
            ASSERT2(
                Path.Pop() != FALSE
                );
            if(Status < WUIMAN_ERRORS)
                {
                WUIMAN_Error(Status, WUIMAN_LastError);
                DEBUG_ERROR("Could not initialize '%s.%s' with value '%s'",
                        Path.AbsolutePath(), (char *)Name, Value);
                DEBUG_ERROR("%s", WUIMAN_LastError);
                }
            }

////////////////////////////////
    // Now load child nodes recursively
    Name    = Keys->FirstKey();
//DEBUG_TRACE("First Child name '%s'.", Name);
    List    = Children();
    if(List)    // if object supports child nodes
        for(; Name != NULL; Name = Keys->NextKey())
            {
            // Read class name, value string
            int ReadStatus = AWuiObject::Database->ReadItem(
                                  Path.AbsolutePath(), Name,
                                        NodeData, MAX_ITEM);
            if(ReadStatus == -1)
                continue;
            if(*Name == '.')  // skip attributes
                continue;
            DEBUG_TRACE("Just read child '%s'='%s'", Name, (char *)NodeData);
            char *Value = strchr(NodeData, ',');
            if(Value)
                *Value++    = '\0';
            AWuiObject *ChildNode = List->Get(Name);
            // if object did not create this child node
            if(ChildNode == 0)
                {   // then create it ourselves
                DEBUG_TRACE("   (had to create '%s'='%s')", Name, (char *)NodeData);
                ChildNode = AWuiObject::NewObject(
                                  TWuiName(NodeData), Name);
                if(!ChildNode)
                    continue;
                List->Insert(ChildNode);
                }
DEBUG_WARNING("  [1c]%s", Path.AbsolutePath());
            Path.Push(Name);
            ChildNode->Load(Path, NULL);
            Path.Pop();
DEBUG_WARNING("  [2c]%s", Path.AbsolutePath());
            }
    delete  Keys;
    return 1;
    }

// Copy() - Provide recursion for copy operation.
void    AWuiObject::Copy(const AWuiObject *Other, int Depth)
    {   MEMBERASSERT();
    // AWuiObject handles copying flag bits
    Flags_  = Other->Flags_;
    TWuiObjects *List       = Attributes();
    TWuiObjects *OtherList  = Other->Attributes();
    // Handle cloning of attributes, if any
    if(IsA(C_TATTRIBUTE) == FALSE
        && List != NULL
        && OtherList != NULL
      )
        {
        int NAttributes = OtherList->NObjects();
        for(int i=0; i < NAttributes; ++i)
            {
            AWuiObject *Other = OtherList->Get(i);
            AWuiObject *This  = List->Get(Other->Name());
            // if we already created this attribute
            if(This != NULL)
                This->Copy(Other,0);  // just copy
            else    // else, must create it
                List->Insert(Other->Clone(Other->Name()));
            }
        }
    // if desired depth is reached, we're done
    if(Depth <= 0)
        return;
    // else, recurse over children
    List        = Children();
    OtherList   = Other->Children();
    if(List != NULL && OtherList != NULL)
        {
        int NAttributes = OtherList->NObjects();
        for(int i=0; i < NAttributes; ++i)
            {
            AWuiObject *Other = OtherList->Get(i);
            AWuiObject *This  = List->Get(Other->Name());
            // if we already created this child
            if(This != NULL)
                This->Copy(Other, Depth-1);  // just copy
            else    // else, must create it
                List->Insert(Other->Clone(Other->Name()), Depth-1);
            }
        }
    }

AWuiObject *AWuiObject::Clone(const TWuiName Type,
                                     const TWuiName NewName)
    {
    ASSERT(AWuiObject::MasterObjects != NULL);
    AWuiObject *Master = AWuiObject::MasterObjects
                                                ->Get(Type);
    char Buffer[TWuiName::MAXNAME];
    if(Master == NULL)
        {
        DEBUG_ERROR("Attempt to create object of undefined "
                "type '%s'", Type.GetName(Buffer));
        return NULL;
        }
    // Found master object, now create new one
    AWuiObject *Object = Master->Clone(NewName);
#if 1
char NBuffer[TWuiName::MAXNAME];
Object->Name().GetName(NBuffer);
DEBUG_TRACE("Cloned name is '%s'", NBuffer);
NewName.GetName(NBuffer);
DEBUG_TRACE("   (Requested name was '%s')", NBuffer);
#endif
    if(Object == NULL)
        {
        DEBUG_ERROR("Could not allocate new object of type '%s'",
                Type.GetName(Buffer));
        return NULL;
        }
    else
        return Object;
    }

AWuiAttribute *AWuiObject::GetAttribute(const TWuiName Name)
    {   MEMBERASSERT();
    TWuiObjects *List = Attributes();
    if(List)
        return (AWuiAttribute *)List->Get(Name);
    else
        return NULL;
    }

TWuiObjects *AWuiObject::Children() const
    {   MEMBERASSERT();
    // default is to not support children
    return NULL;
    }

TWuiObjects *AWuiObject::Attributes() const
    {   MEMBERASSERT();
    // default is to not support attributes
    return NULL;
    }

void    AWuiObject::DefineMethod(const TWuiName MethodName)
    {
    // preconditions
    ASSERT(Methods != NULL);
    ASSERT(Attributes() != NULL);

    // if method is not already on list, create and add it
    AWuiObject  *Method = NULL;
    for(int i=0; i < Methods->NObjects(); ++i)
        if(Methods->Get(i)->Name() == MethodName)
            {
            Method  = Methods->Get(i);
            break;
            }
    if(Method == NULL)
        {
        Method  = new TWuiMethod(MethodName);
        Method->SetFlags(REFERENCE);
        ASSERT(Method != NULL);
        Methods->Insert(Method);
        }

    // Now add to this object's attribute list
    Attributes()->Insert(Method);
    }

